/*
 * Copyright (c) 2020, Oracle and/or its affiliates.
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, are
 * permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice, this list of
 * conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice, this list of
 * conditions and the following disclaimer in the documentation and/or other materials provided
 * with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its contributors may be used to
 * endorse or promote products derived from this software without specific prior written
 * permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package com.oracle.truffle.llvm.runtime.interop;

import com.oracle.truffle.api.dsl.Specialization;
import com.oracle.truffle.llvm.runtime.interop.access.LLVMInteropReadNode;
import com.oracle.truffle.llvm.runtime.interop.access.LLVMInteropType;
import com.oracle.truffle.llvm.runtime.interop.convert.ForeignToLLVM;

/**
 * Node for mapping bitcode function parameters to structs in interop parameters. This is needed due
 * to the fact that a struct (or object) may not be mapped directly 1:1 to function parameters when
 * passed by value to a function.
 */
public abstract class LLVMGetInteropStructParamNode extends LLVMGetInteropParamNode {
    public static LLVMGetInteropStructParamNode create(LLVMInteropType.Struct targetType, int argumentIndex, int memberOffset,
                    ForeignToLLVM.ForeignToLLVMType targetMemberType) {
        return LLVMGetInteropStructParamNodeGen.create(targetType, argumentIndex, memberOffset, targetMemberType);
    }

    private final LLVMInteropType.Struct targetType;
    private final int argumentIndex;
    private final int memberOffset;
    private final ForeignToLLVM.ForeignToLLVMType memberType;

    @Child LLVMInteropReadNode readNode;

    LLVMGetInteropStructParamNode(LLVMInteropType.Struct targetType, int argumentIndex, int memberOffset, ForeignToLLVM.ForeignToLLVMType targetMemberType) {
        this.targetType = targetType;
        this.argumentIndex = argumentIndex;
        this.memberOffset = memberOffset;
        this.memberType = targetMemberType;

        this.readNode = LLVMInteropReadNode.create();
    }

    /**
     * Delegates the lookup of which struct member exists at that particular offset to the
     * InteropReadNode child.
     */
    @Specialization
    Object getParam(Object[] arguments) {
        assert argumentIndex < arguments.length;
        return readNode.execute(targetType, arguments[argumentIndex], memberOffset, this.memberType);
    }
}
