/*
 * Copyright (c) 2018, 2021, Oracle and/or its affiliates.
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, are
 * permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice, this list of
 * conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice, this list of
 * conditions and the following disclaimer in the documentation and/or other materials provided
 * with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its contributors may be used to
 * endorse or promote products derived from this software without specific prior written
 * permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package com.oracle.truffle.llvm.runtime.interop.export;

import com.oracle.truffle.api.dsl.Cached;
import com.oracle.truffle.api.dsl.Fallback;
import com.oracle.truffle.api.dsl.GenerateUncached;
import com.oracle.truffle.api.dsl.Specialization;
import com.oracle.truffle.api.interop.InvalidArrayIndexException;
import com.oracle.truffle.api.interop.UnsupportedMessageException;
import com.oracle.truffle.api.profiles.BranchProfile;
import com.oracle.truffle.llvm.runtime.interop.access.LLVMInteropType;
import com.oracle.truffle.llvm.runtime.nodes.api.LLVMNode;
import com.oracle.truffle.llvm.runtime.pointer.LLVMPointer;

@GenerateUncached
public abstract class LLVMForeignGetIndexPointerNode extends LLVMNode {

    public abstract LLVMPointer execute(LLVMInteropType type, LLVMPointer pointer, long index) throws UnsupportedMessageException, InvalidArrayIndexException;

    @Specialization(guards = "array.elementType == elementType")
    static LLVMPointer doCached(LLVMInteropType.Array array, LLVMPointer pointer, long idx,
                    @Cached("array.elementSize") long elementSize,
                    @Cached("array.elementType") LLVMInteropType elementType,
                    @Cached BranchProfile exception) throws InvalidArrayIndexException {
        if (Long.compareUnsigned(idx, array.length) >= 0) {
            exception.enter();
            throw InvalidArrayIndexException.create(idx);
        }
        return pointer.increment(idx * elementSize).export(elementType);
    }

    @Specialization(replaces = "doCached")
    static LLVMPointer doGeneric(LLVMInteropType.Array array, LLVMPointer pointer, long idx,
                    @Cached BranchProfile exception) throws InvalidArrayIndexException {
        return doCached(array, pointer, idx, array.elementSize, array.elementType, exception);
    }

    /**
     * @param type
     * @param object
     * @param idx
     * @see #execute(LLVMInteropType, LLVMPointer, long)
     */
    @Fallback
    static LLVMPointer doError(LLVMInteropType type, LLVMPointer object, long idx) throws UnsupportedMessageException {
        throw UnsupportedMessageException.create();
    }
}
