/*
 * Copyright (c) 2021, 2021, Oracle and/or its affiliates. All rights reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * The Universal Permissive License (UPL), Version 1.0
 *
 * Subject to the condition set forth below, permission is hereby granted to any
 * person obtaining a copy of this software, associated documentation and/or
 * data (collectively the "Software"), free of charge and under any and all
 * copyright rights in the Software, and any and all patent rights owned or
 * freely licensable by each licensor hereunder covering either (i) the
 * unmodified Software as contributed to or provided by such licensor, or (ii)
 * the Larger Works (as defined below), to deal in both
 *
 * (a) the Software, and
 *
 * (b) any piece of software and/or hardware listed in the lrgrwrks.txt file if
 * one is included with the Software each a "Larger Work" to which the Software
 * is contributed by such licensors),
 *
 * without restriction, including without limitation the rights to copy, create
 * derivative works of, display, perform, and distribute the Software and make,
 * use, sell, offer for sale, import, export, have made, and have sold the
 * Software and the Larger Work(s), and to sublicense the foregoing rights on
 * either these or other terms.
 *
 * This license is subject to the following condition:
 *
 * The above copyright notice and either this complete permission notice or at a
 * minimum a reference to the UPL must be included in all copies or substantial
 * portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package com.oracle.truffle.api.test.polyglot;

import static com.oracle.truffle.api.test.polyglot.HostClassLoadingTest.TEST_REPLACE_CLASS_NAME;
import static com.oracle.truffle.api.test.polyglot.HostClassLoadingTest.createJar;
import static com.oracle.truffle.api.test.polyglot.HostClassLoadingTest.deleteDir;
import static com.oracle.truffle.api.test.polyglot.HostClassLoadingTest.renameHostClass;
import static org.junit.Assert.assertEquals;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;

import org.graalvm.polyglot.Context;
import org.graalvm.polyglot.HostAccess;
import org.graalvm.polyglot.PolyglotException;
import org.graalvm.polyglot.Value;
import org.graalvm.polyglot.io.IOAccess;
import org.junit.Test;

import com.oracle.truffle.tck.tests.TruffleTestAssumptions;

public class SLHostClassLoadingTest {

    @Test
    public void testSLHostClassLoading() throws IOException {
        TruffleTestAssumptions.assumeNotAOT();
        HostAccess accessPolicy = HostAccess.newBuilder(HostAccess.ALL).build();
        try (Context context = Context.newBuilder().allowIO(IOAccess.ALL).allowHostAccess(accessPolicy).allowHostClassLookup(cls -> true).allowHostClassLoading(true).build()) {
            final Class<?> hostClass = HostClassLoadingTestClass1.class;
            Path tempDir = renameHostClass(hostClass, TEST_REPLACE_CLASS_NAME);
            Path jar = createJar(tempDir);
            assertHostClassPath(context, hostClass, jar);
            Files.deleteIfExists(jar);
            deleteDir(tempDir);
        }
    }

    private static void assertHostClassPath(Context context, final Class<?> hostClass, Path jar) {
        String newClassName = hostClass.getPackage().getName() + "." + TEST_REPLACE_CLASS_NAME;

        AbstractPolyglotTest.assertFails(() -> context.eval("sl", "function main() { return java(\"" + newClassName + "\"); }"), PolyglotException.class, (e) -> {
            assertEquals(e.getMessage(), "Access to host class " + newClassName + " is not allowed or does not exist.");
        });

        context.eval("sl", "function main() { return addToHostClassPath(\"" + jar.toString() + "\"); }");

        Value javaType = context.eval("sl", "function main() { return java(\"" + newClassName + "\"); }");
        assertEquals(42, javaType.getMember("staticField").asInt());
        Value javaTypeInstance = javaType.getMember("getInstance").execute();
        assertEquals(42, javaTypeInstance.getMember("testField").asInt());
        assertEquals(42, javaTypeInstance.getMember("testMethod").execute().asInt());
        javaType.putMember("staticField", 43);
        assertEquals(43, javaType.getMember("staticField").asInt());
        javaTypeInstance.putMember("testField", 43);
        assertEquals(43, javaTypeInstance.getMember("testField").asInt());
    }

}
